SystemOrganization addCategory: #Squeak2JS!

OMeta2 subclass: #S2JParser
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Squeak2JS'!

!S2JParser class methodsFor: 'matching' stamp: 'ag 11/18/2010 12:16'!
matchString: aString with: aRule
	
	| matcher result |
	matcher := self matcherOn: aString readStream.
	result := matcher
		apply: #empty;
		apply: aRule.
	matcher apply: #end.
	^ result! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/18/2010 14:54'!
binaryExpression =
	
	(primary | unaryExpression | binaryExpression):receiver binarySelector:message (primary | unaryExpression):argument
	-> [ S2JNode named: 'binaryExpression' with: { receiver. message. argument} ]! !

!S2JParser methodsFor: 'primitive elements' stamp: 'ag 11/18/2010 15:07'!
binarySelector =
	
	(specialCharacter (specialCharacter | empty))
|	('-' (specialCharacter | empty))
|	'|'! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/18/2010 14:22'!
block =
	
	token('[') 
		// arguments
		(((token(':') identifier:argumentName -> [ argumentName ])+:arguments token('|') ) | empty)
		// block contents
		(temporaries:temporaries | empty)
		(statements:statements | empty)
	token(']')
	
	-> [ S2JNode named: 'block' withoutNils: {
			S2JNode named: 'arguments' with: arguments.
			temporaries.
			statements} ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/18/2010 14:15'!
braceExpression =
	
	token('{') ((expression:firstExpression ((token('.') expression:expr -> [ expr ])*:expressions | empty)) | empty) token('}')
	-> [ S2JNode named: 'braceExpression' withoutNils: 
		( expressions ifNil: [ { firstExpression} ] ifNotNil: [ { firstExpression } addAll: expressions ] ) ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/18/2010 14:46'!
expression =
	
	// Variable assignments
	((identifier*:identifiers (token(':=') | token('_')) -> [ identifiers ]):variableNames | empty)
	// Actual part of expression
	(messageExpression | primary | cascadedMessageExpression):expressionBody
	
	// Create an extra node for the assignments, if any
	-> [ S2JNode named: 'expression' withoutNils: { 
		(S2JNode named: 'variableAssignments' withoutNilsNotEmpty: variableNames).
		expressionBody} ]! !

!S2JParser methodsFor: 'primitive elements' stamp: 'ag 11/18/2010 15:05'!
identifier =
	// This is also a unarySelector, the basis for a keyWordSelector, a variable/argument-name
	// This is the only thing, that does not get wrapped into a SJNode (should be clear from context, what this means)
	
	spaces <letter:first (letter | digit)*:rest>:result -> [ result ]! !

!S2JParser methodsFor: 'helpers' stamp: 'ag 11/18/2010 15:16'!
messageExpression =
	
	unaryExpression
|	binaryExpression
	
	// TODO keyword-Expressions!!
	//	|	keywordExpression
! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/18/2010 13:10'!
messagePattern =
	
	// Unary
	identifier:unarySelector -> [ S2JNode named: 'unary' with: { unarySelector } ]
	
	// Binary
|	binarySelector:selector identifier:argumentName -> [ S2JNode named: 'binary' with: { selector. argumentName } ]
	
	// Keyword
	// TODO
! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/18/2010 13:12'!
method =
	
	messagePattern:a1 (temporaries:a2 | empty) (primitiveDeclaration:a3 | empty) (statements:a4 | empty)
	-> [ S2JNode named: 'method' withoutNils: { a1. a2. a3. a4 } ]! !

!S2JParser methodsFor: 'primitive elements' stamp: 'ag 11/18/2010 15:17'!
number =
	
	spaces ^number:number -> [ number ]! !

!S2JParser methodsFor: 'helpers' stamp: 'ag 11/18/2010 15:15'!
primary =
	// something, that evaluates to an object
	
	// TODO constant-array-literal; will need to outsource symbol-, number- and character-literals
	
	identifier // variableName / argumentName
|	number:number -> [ S2JNode named: 'number' with: {number} ]
|	token('#') identifier:symbol -> [ S2JNode named: 'symbol' with: {symbol} ] // TODO symbols can have the form #'abc' and don't have to be identifiers!!
|	token('$') char:char -> [ S2JNode named: 'character' with: {char} ]
|	block
|	braceExpression
|	token('(') expression:expr token(')') -> [ expr ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/18/2010 13:15'!
primitiveDeclaration =
	
	// TODO primitive can also take the form <primitive: methodNameString module: moduleNameString>
	
	token('<') token('primitive:') digits:primitiveNumber token('>')
	-> [ S2JNode named: 'primitive' with: { primitiveNumber } ]! !

!S2JParser methodsFor: 'primitive elements' stamp: 'ag 11/18/2010 14:56'!
specialCharacter =
	
	'+' | '*' | '/' | '\' | '~' | '<' | '>' | '+=' | '@' | '%' | '&' | '?' | '!!' | '`' | ','! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/18/2010 13:55'!
statements =
	
	// Some expressions
	(expression:expr token('.') -> [ expr ] )*:expressions
	
	// The last expression (return statement, not necessary, must not be followed by another statement)
	((token('^') expression:lastExpression (token('.') | empty) ~expression) | empty)
	
	-> [ S2JNode named: 'statements' withoutNils: (expressions add: 
		(lastExpression ifNotNil: [lastExpression name: 'returnExpression'; yourself] ifNil: [ nil ]); yourself) ]
	// Mark the last expression as return-expression
! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/18/2010 13:13'!
temporaries =
	
	token('|') identifier*:variableNames token('|')
	-> [ S2JNode named: 'temporaries' with: variableNames ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/18/2010 14:39'!
unaryExpression =
	
	(primary | unaryExpression):receiver identifier:message
	-> [ S2JNode named: 'unaryExpression' with: { receiver. message} ]! !

OrderedCollection subclass: #S2JNode
	instanceVariableNames: 'name'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Squeak2JS'!

!S2JNode commentStamp: 'ag 11/18/2010 13:09' prior: 0!
This is only a named collection, that is supposed to hold further collections or strings to form a tree!

!S2JNode class methodsFor: 'instance creation' stamp: 'ag 11/18/2010 13:08'!
named: aString with: anOrderedCollection
	
	^ self new
		name: aString;
		addAll: anOrderedCollection;
		yourself! !

!S2JNode class methodsFor: 'instance creation' stamp: 'ag 11/18/2010 13:58'!
named: aString withoutNils: anOrderedCollection
	
	anOrderedCollection ifNil: [ ^ nil ].
	^ self
		named: aString
		with: (anOrderedCollection reject: [ :each | each isNil])! !

!S2JNode class methodsFor: 'instance creation' stamp: 'ag 11/18/2010 13:57'!
named: aString withoutNilsNotEmpty: anOrderedCollection
	"Returns nil, if resulting node would be empty"
	
	| result |
	result := self named: aString withoutNils: anOrderedCollection.
	^ (result isNil | result isEmpty)
		ifTrue: [ nil ]
		ifFalse: [ result ]! !

!S2JNode methodsFor: 'accessors' stamp: 'ag 11/18/2010 13:07'!
name
	
	^ name! !

!S2JNode methodsFor: 'accessors' stamp: 'ag 11/18/2010 13:07'!
name: aString
	
	name := aString! !

!S2JNode methodsFor: 'accessors' stamp: 'ag 11/18/2010 14:35'!
printString
	
	| result first |
	first := true.
	result := name, ' with ('.
	self do: [ :each |
		first ifTrue: [first := false] ifFalse: [result := result, ', '].
		result := result,  each printString].
	^ result, ')'! !
